/*  This file is part of SAIL (https://github.com/smoked-herring/sail)

    Copyright (c) 2020 Dmitry Baryshev

    The MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

#ifndef SAIL_SAIL_DEEP_DIVER_H
#define SAIL_SAIL_DEEP_DIVER_H

#include <stddef.h>

#ifdef SAIL_BUILD
    #include "error.h"
    #include "export.h"
#else
    #include <sail-common/error.h>
    #include <sail-common/export.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

struct sail_context;
struct sail_io;
struct sail_plugin_info;
struct sail_read_options;
struct sail_write_options;

/*
 * Starts reading the specified image file with the specified read options. Pass plugin info if you would like
 * to start reading with a specific codec. If not, just pass NULL. If you do not need specific read options,
 * just pass NULL. Plugin-specific defaults will be used in this case.
 *
 * The read options are deep copied.
 *
 * If read options is NULL, the subsequent calls to sail_read_next_frame() output pixels in BPP32-RGBA
 * pixel format for image formats with transparency support and BPP24-RGB otherwise.
 *
 * Typical usage: sail_start_reading_file_with_options() ->
 *                sail_read_next_frame()                 ->
 *                sail_stop_reading().
 *
 * Or:            sail_plugin_info_from_extension()      ->
 *                sail_start_reading_file_with_options() ->
 *                sail_read_next_frame()                 ->
 *                sail_stop_reading().
 *
 * STATE explanation: Passes the address of a local void* pointer. SAIL will store an internal state
 * in it and destroy it in sail_stop_reading(). States must be used per image. DO NOT use the same state
 * to start reading multiple images at the same time.
 *
 * Returns 0 on success or sail_error_t on error.
 */
SAIL_EXPORT sail_error_t sail_start_reading_file_with_options(const char *path, struct sail_context *context,
                                                              const struct sail_plugin_info *plugin_info,
                                                              const struct sail_read_options *read_options, void **state);

/*
 * Starts reading the specified memory buffer with the specified read options. If you do not need specific read options,
 * just pass NULL. Plugin-specific defaults will be used in this case.
 *
 * The read options are deep copied.
 *
 * If read options is NULL, the subsequent calls to sail_read_next_frame() output pixels in BPP32-RGBA
 * pixel format for image formats with transparency support and BPP24-RGB otherwise.
 *
 * Typical usage: sail_plugin_info_from_extension()     ->
 *                sail_start_reading_mem_with_options() ->
 *                sail_read_next_frame()                ->
 *                sail_stop_reading().
 *
 * STATE explanation: Passes the address of a local void* pointer. SAIL will store an internal state
 * in it and destroy it in sail_stop_reading(). States must be used per image. DO NOT use the same state
 * to start reading multiple images at the same time.
 *
 * Returns 0 on success or sail_error_t on error.
 */
SAIL_EXPORT sail_error_t sail_start_reading_mem_with_options(const void *buffer, size_t buffer_length, struct sail_context *context,
                                                             const struct sail_plugin_info *plugin_info,
                                                             const struct sail_read_options *read_options, void **state);

/*
 * Starts writing the specified image file with the specified write options. Pass plugin info if you would like
 * to start writing with a specific codec. If not, just pass NULL. If you do not need specific write options,
 * just pass NULL. Plugin-specific defaults will be used in this case.
 *
 * The write options are deep copied.
 *
 * If write options is NULL, the subsequent calls to sail_write_next_frame() output pixels in pixel format
 * as specified in sail_write_features.preferred_output_pixel_format.
 *
 * Typical usage: sail_start_writing_file_with_options() ->
 *                sail_write_next_frame()                ->
 *                sail_stop_writing().
 *
 * Or:            sail_plugin_info_from_extension()      ->
 *                sail_start_writing_file_with_options() ->
 *                sail_write_next_frame()                ->
 *                sail_stop_writing().
 *
 * STATE explanation: Passes the address of a local void* pointer. SAIL will store an internal state
 * in it and destroy it in sail_stop_writing(). States must be used per image. DO NOT use the same state
 * to start writing multiple images at the same time.
 *
 * Returns 0 on success or sail_error_t on error.
 */
SAIL_EXPORT sail_error_t sail_start_writing_file_with_options(const char *path, struct sail_context *context,
                                                              const struct sail_plugin_info *plugin_info,
                                                              const struct sail_write_options *write_options, void **state);

/*
 * Starts writing the specified memory buffer with the specified write options. If you do not need specific
 * write options, just pass NULL. Plugin-specific defaults will be used in this case.
 *
 * The write options are deep copied.
 *
 * If write options is NULL, the subsequent calls to sail_write_next_frame() output pixels in pixel format
 * as specified in sail_write_features.preferred_output_pixel_format.
 *
 * Typical usage: sail_plugin_info_from_extension()     ->
 *                sail_start_writing_mem_with_options() ->
 *                sail_write_next_frame()               ->
 *                sail_stop_writing().
 *
 * STATE explanation: Passes the address of a local void* pointer. SAIL will store an internal state
 * in it and destroy it in sail_stop_writing. States must be used per image. DO NOT use the same state
 * to start writing multiple images at the same time.
 *
 * Returns 0 on success or sail_error_t on error.
 */
SAIL_EXPORT sail_error_t sail_start_writing_mem_with_options(void *buffer, size_t buffer_length, struct sail_context *context,
                                                             const struct sail_plugin_info *plugin_info,
                                                             const struct sail_write_options *write_options, void **state);


/*
 * Stops writing the file started by sail_start_writing_file() and brothers. Assigns the number of bytes written.
 * Does nothing if the state is NULL.
 *
 * It is essential to always stop writing to free memory resources. Failure to do so will lead to memory leaks.
 *
 * Returns 0 on success or sail_error_t on error.
 */
SAIL_EXPORT sail_error_t sail_stop_writing_with_written(void *state, size_t *written);

/* extern "C" */
#ifdef __cplusplus
}
#endif

#endif
